/* XVCL - the xview classes

   XVCL might one day be a hopefully useful class library
   for xview programming in C++. Just now, it slowly grows
   as I write programs, so many many things are missing.

   Copyright (C) 1998 Frank Rustemeyer.
   This file is part of the Emperor package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   If you have changes and you want them to be included in a future
   release of Emperor, please e-mail them to mambres@mud.tap.de.

   This file implements xview classes for windows, panels, buttons and so on.

   98/11/30, Frank Rustemeyer <mambres@mud.tap.de>: Released.
*/

#include <xview/xview.h>
//...more xview...h's are included in xvcl.h
#include <xview/cms.h>
#include <X11/Xlib.h>
#include <xview/notice.h>
#include <xview/file_chsr.h>
#include <xview/scrollbar.h>
#include <xview/notify.h>
#include "xvcl.h"
#include "xpm.h"

distributor *d;

// ********************************************************* xobj *****

void xobj::show(int i)   // display the object
{   xv_set(get_handle(), XV_SHOW, i, END);
    }

int xobj::shown(void)
{   return (int) xv_get(get_handle(), XV_SHOW);
    }

// ******************************************************** window ****

void window::create_frame(int x, int y, char *label)
{   Frame hW;

    hW = (Frame) xv_create(0, FRAME,
                 FRAME_LABEL, label,
                 XV_WIDTH, x, XV_HEIGHT, y,
                 END);

    set_handle(hW);
    }

void window::set_min_size(int x, int y)
{   xv_set(get_handle(),
           FRAME_MIN_SIZE, x, y,
           END );
    }

void window::set_max_size(int x, int y)
{   xv_set(get_handle(),
           FRAME_MAX_SIZE, x, y,
           END);
    }

void window::notify_destroy(void)
{   xv_set(get_handle(),
           FRAME_DONE_PROC, centraldestroy,
           END);
    }

// ******************************************************* panel ******

panel::panel(class window *w)
{   set_handle(
        xv_create(w->get_handle(), PANEL, END)
        );
    owner = w;
    }

panel::set_height(int h)
{   xv_set(get_handle(), 
           XV_HEIGHT, h,
           END);
    }
    
panel::set_width(int w)
{   xv_set(get_handle(),
	   XV_WIDTH, w,
           END);
    }

void panel::button_pressed(Xv_object button)
{   owner->button_pressed(button);
    }

void panel::checkbox_selected(Xv_object item, int value)
{   owner->checkbox_selected(item, value);
    }

void panel::menu_selected(char *item)
{   owner->menu_selected(item);
    }

Display * panel::query_display(void)
{   return (Display *) xv_get(get_handle(), XV_DISPLAY);
    }

// ****************************************************** canvas ******

canvas::canvas(class window *w, int x, int y, 
               int ncolours, Xv_singlecolor *cols)
{   Canvas hC;
    cms = (Cms) xv_create(0, CMS,
           CMS_SIZE, ncolours,
           CMS_COLORS, cols,
           END );
    coltable = (unsigned long *) xv_get(cms, CMS_INDEX_TABLE);
    hC = (Canvas) xv_create(w->get_handle(), CANVAS,
          CANVAS_X_PAINT_WINDOW, TRUE,
          CANVAS_REPAINT_PROC, centralpaint,
          CANVAS_FIXED_IMAGE, FALSE,
          XV_X, x, XV_Y, y,
          WIN_CMS, cms,
          END );
    owner = w;
    set_handle(hC);
    }

canvas::canvas(class window *w, int x, int y, Cms c)
{   Canvas hC;
    hC = (Canvas) xv_create(w->get_handle(), CANVAS,
          CANVAS_X_PAINT_WINDOW, TRUE,
          CANVAS_REPAINT_PROC, centralpaint,
          CANVAS_FIXED_IMAGE, FALSE,
          XV_X, x, XV_Y, y,
          WIN_CMS, c,
          END );
    owner = w;
    cms = c;
    coltable = (unsigned long *) xv_get(cms, CMS_INDEX_TABLE);
    set_handle(hC);
    }

canvas::set_fixed_size(int x, int y)
{   xv_set(get_handle(),
          CANVAS_AUTO_SHRINK, FALSE,
          CANVAS_AUTO_EXPAND, FALSE,
          CANVAS_WIDTH, x,
          CANVAS_HEIGHT, y,
          END);
    }

canvas::add_scrollbars(void)
{   s_ver = (Scrollbar) xv_create(get_handle(), SCROLLBAR,
                                  SCROLLBAR_DIRECTION, SCROLLBAR_VERTICAL,
                                  END);
    s_hor = (Scrollbar) xv_create(get_handle(), SCROLLBAR,
                                  SCROLLBAR_DIRECTION, SCROLLBAR_HORIZONTAL,
                                  END);
    }

void canvas::scrollto(int x, int y)
{   xv_set(s_hor, SCROLLBAR_VIEW_START, x, END);
    xv_set(s_ver, SCROLLBAR_VIEW_START, y, END);
    }

void canvas::centre(int x, int y)
{   int sx, sy, wx, wy, ox, oy;
    wx = xv_get(s_hor, SCROLLBAR_VIEW_LENGTH);
    wy = xv_get(s_ver, SCROLLBAR_VIEW_LENGTH);
    ox = xv_get(s_hor, SCROLLBAR_OBJECT_LENGTH);
    oy = xv_get(s_ver, SCROLLBAR_OBJECT_LENGTH);

    sx = x - (wx / 2);
    if (sx < 0) sx = 0;
    sy = y - (wy / 2);
    if (sy < 0) sy = 0;

    if (sx > ox - wx) sx = ox - wx;
    if (sy > oy - wy) sy = oy - wy;
    scrollto(sx, sy);
    }

int canvas::visible(int x, int y)
{   int minx, miny;
    minx = xv_get(s_hor, SCROLLBAR_VIEW_START);
    miny = xv_get(s_ver, SCROLLBAR_VIEW_START);
    if (x < minx) return 0;
    if (y < miny) return 0;
    if (x > minx + xv_get(s_hor, SCROLLBAR_VIEW_LENGTH)) return 0;
    if (y > miny + xv_get(s_ver, SCROLLBAR_VIEW_LENGTH)) return 0;
    return 1;
    }

void canvas::mouse_events(void)
{   xv_set(paint_window(),
           WIN_EVENT_PROC, centralevent,
           WIN_CONSUME_EVENTS, WIN_MOUSE_BUTTONS, END,
           END );
    }

void canvas::keyboard_events(void)
{   xv_set(paint_window(),
           WIN_EVENT_PROC, centralevent,
           WIN_CONSUME_EVENTS, WIN_ASCII_EVENTS, END,
           END );
    }

Xv_object canvas::paint_window(void)
{   return canvas_paint_window(get_handle());
    }

Display * canvas::query_display(void)
{   return (Display *) xv_get(paint_window(), XV_DISPLAY);
    }

void canvas::invalidate(void)
{   owner->paint(get_handle(), paint_window(), query_display(),
                 xv_get(paint_window(), XV_XID), (Xv_xrectlist *) NULL);
    }

// ****************************************************** dialog ******

dialog::dialog(void) : window()
{ 
    }

int dialog::run(void) // execute the dialog
{   xv_window_loop(get_handle());
    return retval;
    }

void dialog::enddialog(int returnval)
{   retval = returnval;
    xv_window_return(0);
    }

dialog::~dialog()
{   xv_destroy_safe(get_handle());
    d->removewindow(get_handle());
      // removes this window from the distributor's list of active windows
    }

void dialog::destroy(void)
{   xv_window_return(0);
    }

// ***************************************************** panelitem ****

panelitem::~panelitem(void)
{   d->removewindow(get_handle());
    }

// ****************************************************** button ******

button::button(panel *p, char * text)
  // creates a button in the Panel "p", labels it "text" and
  // sends button messages to o->react(button handle)
{   Panel_item hB;
    hB =      xv_create(p->get_handle(), PANEL_BUTTON,
                        PANEL_LABEL_STRING, text,
                        PANEL_NOTIFY_PROC, centralbuttonpressed,
                        END );
    owner = p;
    set_handle(hB);
    d->addwindow(this, hB);
    }

void button::react(void)
{   owner->button_pressed(get_handle());
    }

void button::add_menu(class menu *m)
{   xv_set(get_handle(), PANEL_ITEM_MENU, m->get_handle(), END);
    }

void button::enable(int state)
{   if (!state) xv_set(get_handle(), PANEL_INACTIVE, TRUE, END);
          else  xv_set(get_handle(), PANEL_INACTIVE, FALSE, END);
    }

// ****************************************************** gauge ******

gauge::gauge(panel *p)
{   Panel_item hG;
    hG = xv_create(p->get_handle(), PANEL_GAUGE,
                   PANEL_SHOW_RANGE, TRUE,
                   END);
    owner = p;
    set_handle(hG);
    }

void gauge::set_value(int v)
{   xv_set(get_handle(), PANEL_VALUE, v, END);
    }

void gauge::set_width(int w)
{   xv_set(get_handle(), PANEL_GAUGE_WIDTH, w, END);
    }

void gauge::set_range(int min, int max)
{   xv_set(get_handle(), 
           PANEL_MIN_VALUE, min,
           PANEL_MAX_VALUE, max,
           END);
    }

// ******************************************************* menu ******

menu::menu(window *o, int size, char **entries)
{   Menu hM;
    Menu_item hMI;
    int i;

    hM = (Menu) xv_create(0, MENU, END);
 
    set_handle(hM);
    owner = o;
    d->addwindow(this, hM);

    for (i = 0; i < size; i++)
    {  hMI = xv_create(0, MENUITEM,
                       MENU_STRING, entries[i],
                       MENU_NOTIFY_PROC, centralmenu,
                       MENU_RELEASE,
                       END);
       xv_set(hM, MENU_APPEND_ITEM, hMI, END);
       }
    }
                          
menu::~menu(void)
{   d->removewindow(get_handle());
    }

void menu::selected(Menu_item item)
{   owner->menu_selected((char *) xv_get(item, MENU_STRING));
    }

// ***************************************************** checkbox ****

checkbox::checkbox(panel *p, char * text, int val)
   // creates a checkbox in the Panel "panel"
{   Panel_item hCB;
    hCB =     xv_create(p->get_handle(), PANEL_CHECK_BOX,
                        PANEL_CHOICE_STRINGS, text, END, 
                        PANEL_NOTIFY_PROC, centralselected,
                        PANEL_VALUE, val,
                        END);
    owner = p;
    set_handle(hCB);
    d->addwindow(this, hCB);
    }

void checkbox::selected(int val)
{   owner->checkbox_selected(get_handle(), val);
    }

void checkbox::set_value(int val)
{   xv_set(get_handle(), 
           PANEL_VALUE, val,
           END);
    }

int checkbox::query_value(void)
{   return (int) xv_get(get_handle(), PANEL_VALUE);
    }

// ***************************************************** textbox *****

textbox::textbox(panel *p, char *text, char *val, int maxlen)
{   Panel_item hTB;
    hTB =     xv_create(p->get_handle(), PANEL_TEXT,
                        PANEL_LABEL_STRING, text,
                        PANEL_VALUE, val,
                        PANEL_VALUE_STORED_LENGTH, maxlen,
                        PANEL_VALUE_DISPLAY_LENGTH, maxlen,
                        END);
    owner = p;
    set_handle(hTB);
    d->addwindow(this, hTB);
    }

void textbox::set_value(char *val)
{   xv_set(get_handle(),
           PANEL_VALUE, val,
           END);
    }

char *textbox::query_value(void)
{   return (char *) xv_get(get_handle(), PANEL_VALUE);
    }
  
// **************************************************** numeric ******

numeric::numeric(panel *p, char *text, int val, int min, int max)
{   Panel_item hN;
    hN =      xv_create(p->get_handle(), PANEL_NUMERIC_TEXT,
                        PANEL_LABEL_STRING, text,
                        PANEL_VALUE, val,
                        PANEL_MIN_VALUE, min,
                        PANEL_MAX_VALUE, max,
                        END);
    owner = p;
    set_handle(hN);
    }

void numeric::set_value(int val)
{   xv_set(get_handle(),
           PANEL_VALUE, val,
           END);
    }

int numeric::query_value(void)
{   return (int) xv_get(get_handle(), PANEL_VALUE);
    }

// ****************************************************** textview ***

textview::textview(panel *p, char *text)
{   Panel_item hTV;
    hTV = xv_create(p->get_handle(), PANEL_MESSAGE,
                    PANEL_LABEL_STRING, text,
                    END);
    owner = p;
    set_handle(hTV);
    }

void textview::set_value(char *val)
{   xv_set(get_handle(),
           PANEL_LABEL_STRING, val,
           END);
    }

// ****************************************************** imageview **

imageview::imageview(panel *p, char **img)
{   Panel_item hIV;
    pixmap *pm;
    Server_image si;

    pm = new pixmap(p->query_display(), img);
    si = (Server_image) xv_create(0, SERVER_IMAGE,
                                  SERVER_IMAGE_PIXMAP, pm->query_pixmap(),
                                  END);
    delete pm;
 
    hIV = xv_create(p->get_handle(), PANEL_MESSAGE,
                    PANEL_LABEL_IMAGE, si,
                    END);
    owner = p;
    set_handle(hIV);
    }

// ****************************************************** listbox ****

listbox::listbox(panel *p, char *title, int height)
{   Panel_item hL;
    hL = xv_create(p->get_handle(), PANEL_LIST,
                   PANEL_LIST_DISPLAY_ROWS, height,
                   END);
    if (title) xv_set(hL, PANEL_LIST_TITLE, title, END);
    owner = p;
    set_handle(hL);
    }

void listbox::add_string(char *s)
{   int next;
    next = xv_get(get_handle(), PANEL_LIST_NROWS);
    xv_set(get_handle(), PANEL_LIST_INSERT, next, END);
    xv_set(get_handle(), PANEL_LIST_STRING, next, s, END);
    }

int listbox::get_value(void)
{   return (int) xv_get(get_handle(), PANEL_LIST_FIRST_SELECTED);
    }

void listbox::set_value(int n)
{   xv_set(get_handle(), 
           PANEL_LIST_SELECT, n, TRUE,
           END);
    }

// *********************************************** messagebox ********

messagebox::messagebox(xobj *owner, char *text)
{   set_handle(
         xv_create(owner->get_handle(), NOTICE,
                   NOTICE_MESSAGE_STRING, text,
                   XV_SHOW, TRUE,
                   END)
         );
    }

// ********************************************** yesnobox ***********

yesnobox::yesnobox(xobj *owner, char *text) 
{   val = 0;
    set_handle(
         xv_create(owner->get_handle(), NOTICE,
                   NOTICE_MESSAGE_STRING, text,
                   NOTICE_BUTTON_YES, "Ja",
                   NOTICE_BUTTON_NO, "Nein",
                   NOTICE_STATUS, &val,
                   XV_SHOW, TRUE,
                   END)
         );
    }

int yesnobox::ok(void)
{   return (val == NOTICE_YES); 
    }

// ************************************************ infobox **********

infobox::infobox(xobj *owner, char *text)
{   panel *p;
    textview *tv;
    set_handle(
         xv_create(owner->get_handle(), FRAME,
                     FRAME_LABEL, "Info",
                     XV_WIDTH, 220,
                     END));
    p = new panel(this);
    p->set_height(25);
    tv = new textview(p, text);
    window_fit(get_handle());
    }
     
// ********************************************** filechooser ********

filechooser::filechooser(window *o, int typ, char * filter)
{   File_chooser hFC;

    owner = o;
    if (typ == FILE_OPEN)
         hFC = xv_create(owner->get_handle(), FILE_CHOOSER_OPEN_DIALOG,
                    XV_LABEL, "Datei ffnen",
                    FILE_CHOOSER_NOTIFY_FUNC, centralfileopen,
                    FILE_CHOOSER_FILTER_STRING, filter,
                    END);
    else
         hFC = xv_create(owner->get_handle(), FILE_CHOOSER_SAVE_DIALOG,
                    XV_LABEL, "Datei speichern",
                    FILE_CHOOSER_NOTIFY_FUNC, centralfilesave,
                    FILE_CHOOSER_DOC_NAME, filter,
                    END);

    set_handle(hFC);
    d->addwindow(this, hFC);
    }

void filechooser::activate(void)
{   xv_set(get_handle(), XV_SHOW, TRUE, END);
    }

int filechooser::file_name(char * path, char *file)
{   owner->file_open(path);
    return XV_OK;
    }

int filechooser::file_save(char *path)
{   owner->file_save(path);
    return XV_OK;
    }

// *************** The Distributor object ****************************

// this object keeps a list of all active window objects and their
// window handle. When the centralwndproc receives a windows message
// intended for the window with a given handle, it can ask the
// the distributor for the associated window object and then call
// the virtual wndproc in that window object to process the message

distributor::distributor(void)
{   int i;
    warray = new xobj * [100];  // list of window pointers
    harray = new Xv_object   [100];  // list of associated handles
    for (i=0; i<100; i++)     // clear array
    {   warray[i] = (xobj *) NULL;
	harray[i] = (Xv_object) 0;
	}
    nwindows = 0;  // number of windows registered
    }

distributor::~distributor(void)
{   delete [] warray;
    delete [] harray;
    }

int distributor::addwindow(xobj *w, Xv_object h)
{   if (nwindows>99) return 0; // improbable
    warray[nwindows] = w;
    harray[nwindows++] = h;
    return 1;
    }

int distributor::removewindow(Xv_object h)
  // remove a window from the window list when it is destroyed
{   int i=0;
    while(harray[i] != h)   // find window in the list
       if (++i == nwindows) return 0;
    for( ; i<nwindows; i++)  // shift all following entries up
    {   warray[i] = warray[i+1];
	harray[i] = harray[i+1];
	}
    warray[nwindows] = NULL; //set last entry to zero
    harray[nwindows] = 0;
    nwindows--;    // decrease number of windows in list 
    return 1;
    } 

xobj * distributor::findobject(Xv_object h)
  // returns the window object managing the window with handle hW
{   int i=0;
    if (!nwindows) return (xobj *) NULL;  // no windows in the list
    while(harray[i]!=h)    // find object in the list
       if(++i == nwindows) return (xobj *) NULL; // object not in list
    return warray[i];  // return object
    }


//********************************************************* Callbacks ******
//  The following functions distribute windows messages.


void centralpaint(Canvas c, Xv_Window x, Display *dsp, Window wnd, Xv_xrectlist *xr)
{   window *w;
    w = (window *) d->findobject((Xv_object) wnd);  
         // get object pointer from distributor
    if (!w) return;
    w->paint(c, x, dsp, wnd, xr);
    }

void centralresize(Canvas c, int x, int y)
{   window *w;
    w = (window *) d->findobject(xv_get(canvas_paint_window(c), XV_XID));
    if (!w) return;
    w->resize(x, y);
    }

void centraldestroy(Frame f)
{   xobj *w;
    w = d->findobject((Xv_object) f);
    if (!w) return;
    w->destroy();
    }

void centralbuttonpressed(Panel_item item)
{   button *b;
    b = (button *) d->findobject((Xv_object) item);
    if (!b) return;
    b->react();
    }

void centralevent(Xv_Window wnd, Event *e, Notify_arg arg, Notify_event_type t)
{   window *w;
    w = (window *) d->findobject(wnd);
    if (!w) return;
    w->event(wnd, e, arg, t);
    }

void centralselected(Panel_item item, int val)
{   checkbox *cb;
    cb = (checkbox *) d->findobject((Xv_object) item);
    if (!cb) return;
    cb->selected(val);
    }

void centralmenu(Menu men, Menu_item item)
{   menu *m;
    m = (menu *) d->findobject((Xv_object) men);
    if (!m) return;
    m->selected(item);
    }

int centralfileopen(Xv_object hFC, char *path, char *file)
{   filechooser *fc;   
    fc = (filechooser *) d->findobject((Xv_object) hFC);
    if (!fc) return XV_ERROR;
    return fc->file_name(path, file);
    }

int centralfilesave(Xv_object hFC, char *path)
{   filechooser *fc;
    fc = (filechooser *) d->findobject((Xv_object) hFC);
    if (!fc) return XV_ERROR;
    return fc->file_save(path);
    }

Notify_value centralinput(Notify_client c, int fd)
{   window *w;
    w = (window *) d->findobject(c);
    if (!w) return NOTIFY_DONE;
    return w->input_pending(fd);
    }
